/************************************************************
 * Firmware for SA1200P's piggybacked EPS8266
 * 
 * Author: Bob Fisch
 *         based on Guy Weiler's exellent work 
 *         http://www.weigu.lu/microcontroller/co2_sa1200p_hacks/index.html
 * 
 * /!\  DON'T FORGET TO UPLOAD SPIFFS DATA FOLDER  /!\
 *  
 * The plugin for Arduino IDA can be found here:
 * https://github.com/esp8266/arduino-esp8266fs-plugin/releases/tag/0.5.0
 */

//#define DEBUG

// code is split up into different files
#include "mywifi.h"
#include "myspiffs.h"
#include "mywebserver.h"
#include "mymqtt.h"
#include "myco2.h"
#include "myhumtemp.h"

// default login data
String wifi_ssid = generateSSID();
String wifi_pass = "123456789";

// mqtt broker settings
String mqtt_server  = "";
String mqtt_port    = "";
String mqtt_user    = "";
String mqtt_pass    = "";
String mqtt_clientid= "";
String mqtt_ssl     = "";
int    mqtt_connect_count = 0;

// default access control password
String passwd = "default";

// wifi & mqtt client
WiFiClient *espClient;
PubSubClient *client;

// needed for scheduling event
long lastMsg100ms = 0;
long lastMsg1s = 0;
long lastMsg3s = 0;
long lastMsg1m = 0;

// heartbeat LED
#define LED_PIN D4

// define different modes
#define MODE_CONFIG  0
#define MODE_RUN     1
int mode = MODE_CONFIG;

// web server object
ESP8266WebServer server(80);

// needed for reading the CO2 sensor
const byte CO2_RX = 14;  // D5
int co2;
SoftwareSerial SerialCO2(CO2_RX); // RX

// needed for reading hum & temp
const byte SDA_PIN = 4; // D2
const byte SCL_PIN = 5; // D1
byte event_counter = 0;
float sht21_value;
float hum, temp;
const byte buffer_bytes = 8;
volatile byte bit_buffer[buffer_bytes];
volatile byte bit_counter = 0;  

void setup() {
  // first let's start the console
  Serial.begin(74880);
  Serial.println("Starting ...");

  // next, start SPIFFS
  Serial.println("Setting up SPIFFS ...");
  SPIFFS.begin();

  // get boot count resetter from SPIFFS
  String bcr = getProperty("BCR");
  // increment by one
  setProperty("BCR",bcr+".");
  Serial.println("BCR is now "+(String)bcr.length());

  // get WIFI setting from SPISS
  String fSSID = getProperty("SSID");
  String fPASS = getProperty("PASS");

  // get MQTT sttings from SPIFFS
  mqtt_server  = getProperty("SERVER");
  mqtt_port    = getProperty("PORT");
  mqtt_user    = getProperty("USERNAME");
  mqtt_pass    = getProperty("PASSWORD");
  mqtt_clientid= getProperty("CLIENT");
  mqtt_ssl     = getProperty("SSL");

  // get the access control password from SPIFFS
  String tpwd = getProperty("PASSWD");
  // set to default if empty
  if(tpwd!="") passwd=tpwd;

  // create the wifi client based on the security settings
  espClient = (mqtt_ssl=="0") ? new WiFiClient() : new WiFiClientSecure();
  client = new PubSubClient(*espClient);
  if(mqtt_ssl=="1")
  {
    WiFiClientSecure *secure = (WiFiClientSecure*)espClient;
    // do not validate the server certificate
    // (ESP does not have the CA loaded!)
    secure->setInsecure();
  }

  // define inputs and outputs
  pinMode(LED_PIN, OUTPUT);
  pinMode(SDA_PIN,INPUT);
  pinMode(SCL_PIN,INPUT);

  // entre running mode
  // if SSID & PASS in settings file
  // and BCR did not yet reach 5
  if(fSSID!="" && fPASS!="" && bcr.length()<5)
  {
    mode = MODE_RUN;
    // use informatio read from SPIFFS
    wifi_ssid = fSSID;
    wifi_pass = fPASS;
  }

  Serial.println("Using SSID = "+wifi_ssid);
  Serial.println("Using PASS = "+wifi_pass);
  
  if(mode==MODE_CONFIG)
  {
    // open an AP for configuration
    Serial.println("Setting up WiFi AP ...");
    WiFi.softAP(wifi_ssid.c_str(), wifi_pass.c_str());
    IPAddress myIP = WiFi.softAPIP();
    Serial.print("IP address: ");
    Serial.println(WiFi.softAPIP());
  }
  else
  {
    // connect to existing WIFI network
    connectWIFI();
    
    // init CO2 sensor
    SerialCO2.begin(9600);
    delay(500);    
    attachInterrupt(digitalPinToInterrupt(SCL_PIN), isr_scl, RISING);
    noInterrupts();  
    clear_buffer(buffer_bytes);
    bit_counter = 0;  
    interrupts(); // allow interrupts
    
    client->loop();

    // if we reach this line, we can clear BRC
    // this also means, that if you want to reset the device,
    // switch on/off power very quickly!
    setProperty("BCR","");
  }  

  // setup webserver
  server.on("/",handleRoot);
  server.on("/co2",handleCO2);
  server.on("/temp",handleTemp);
  server.on("/hum",handleHum);
  server.onNotFound(handleNotFound);
  server.begin();
}  

void loop() {    
  long now = millis();

  // MQTT (retry once minute to reconnect
  if (now - lastMsg1m > 60*1000) {
    lastMsg1m = now;
    if(!client->connected())
      mqtt_connect_count=0;
  }
  
  if(mode==MODE_RUN)
  {
    // reconnect if no more WIFI present
    if (WiFi.status() != WL_CONNECTED) 
      connectWIFI();
    
    // reconnect client if not yet connected
    // but only if all settings are filled 
    if (WiFi.status() == WL_CONNECTED &&
        !client->connected() && 
        mqtt_server!="" && mqtt_port!="" && 
        mqtt_user!="" && mqtt_pass!="" && 
        mqtt_clientid!="") 
      connectMQTT();

    // make client live
    client->loop();
  
    // heartbeat
    if (now - lastMsg1s > 1000) {
      lastMsg1s = now;
      digitalWrite(LED_PIN,!digitalRead(LED_PIN));
    }

    // CO2 can't be faster than 3s
    if (now - lastMsg3s > 3000) {
      doCO2();
      lastMsg3s = now;
    }

    // hum & temp needs to be faster
    // but not too fast, 
    // because we sniff the bus
    if (now - lastMsg100ms > 100) {
      doHumTemp();
      lastMsg100ms = now;
    }
  }
  
  // handle web server clients
  server.handleClient();

  // pass control to other tasks
  yield();
}
